<?php
/*  Copyright 2008  Michael J. Walker  (email : mike@englishmike.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

add_action('publish_page', 'az_cache_post_changed');
add_action('publish_post', 'az_cache_post_changed');
add_action('save_post', 'az_cache_post_changed');
add_action('deleted_post', 'az_cache_post_changed');
add_action('wp_update_post', 'az_cache_post_changed');
add_action('wp_insert_post', 'az_cache_post_changed');
add_action('check_ajax_referer', 'az_cache_check_custom');

//add_action('wp_ajax_addmeta', 'az_cache_check');
//add_action('edit_post', 'az_cache_check');
//add_action('init', 'az_cache_check');

function az_cache_get($index, $pageno) {
	return new az_index_cache($index, $pageno);
}

function az_cache_post_changed($postid) {
    az_trace("az_cache_post_changed:".current_filter()." - postid = $postid");
	az_add_dirty_post($postid);
}

function az_cache_check_custom($action) {
	global $wpdb;
    az_trace("az_cache_check:".current_filter()." - action = $action");
    $action = $_REQUEST['action'];
    if ($action == 'add-meta') {
        $postid = $_REQUEST['post_id'];
        if (empty($postid)) {
        	$meta = $_REQUEST['meta'];
            $idmeta = current(array_keys($meta));
        }
    } else if ($action == 'delete-meta') {
        $idmeta = $_REQUEST['id'];
    }
    
    if (!empty($idmeta)) {
        $postid = $wpdb->get_var("SELECT post_id FROM $wpdb->postmeta WHERE meta_id = '$idmeta'");
    }
    
    if (!empty($postid)) {
        az_trace("custom field modified - postid: $postid");
        az_add_dirty_post($postid);
    }
}

function az_get_dirty_posts() {
    return unserialize(get_option("az_cache_dirty"));	
}

function az_set_dirty_posts($dirtyposts) {
	update_option("az_cache_dirty", serialize($dirtyposts));
}

function az_add_dirty_post($postid) {
	$indexes = az_get_indexes("idindex");
	// Acquire the cache semaphore while we are updating the dirty list.
	az_acquire_cache_mutex();
    $dirtyposts = az_get_dirty_posts();
    // Only process the dirty post if its the same post as before if
    // a quarter of a second or more has elapse, otherwise we will
    // process the same changes multiple time. 
    $elapsed = ($postid == $dirtyposts[0][0]) ? microtime(true) - $dirtyposts[0][1] : 1;
    if ($elapsed > 0.25) {
        // Add the dirty post to all the indexes since we have no idea which
        // posts are displayed in which indexes.
        foreach ($indexes as $index) {
 	        if (!(is_array($dirtyposts[$index->idindex]) && in_array(intval($postid), $dirtyposts[$index->idindex]))) {
                $dirtyposts[$index->idindex][] = intval($postid);
            }
            //az_trace("dirty: id = ".$index->idindex.": ".(implode(",", $dirtyposts[$index->idindex])));
	   }
	}
	// Save the postid and timestamp for checking next time we hook a function.
    $dirtyposts[0][0] = $postid;
    $dirtyposts[0][1] = microtime(true);
    az_set_dirty_posts($dirtyposts);
    // Release the semaphore now that we are done.
    az_release_cache_mutex();
}

function az_reset_dirty_posts($dirtyposts, $idindex) {
	$dirtyposts[$idindex] = array();
    az_set_dirty_posts($dirtyposts);
}

function az_cache_index_removed($idindex) {
    az_acquire_cache_mutex();
    $dirtyposts = az_get_dirty_posts();
    unset($dirtyposts[$idindex]);
    az_set_dirty_posts($dirtyposts);
    az_release_cache_mutex();
}

class az_index_cache {
	var $index;          // Current index
	var $pageno;         // Current page number
    var $pagecount;      // Number of pages in the index
	var $items;          // Array of items;
    var $itemcache;      // Cache of item ids stored in database
	var $alphalinks;     // Array of alphabetical links
	var $ignorechars;    // Characters to ignore in a sort
	var $is_multipage;   // True if a multipage index
    var $has_alphalinks; // True if alphalinks being used
    var $cache_disabled; // True if caching is disabled for this index
    
	function az_index_cache($index, $pageno) {
		$time_start = microtime(true);
        
		$this->cache_disabled = az_is_set($index->options, 'disable-cache');
        $this->has_alphalinks = az_is_set($index->options, 'alpha-links');
        $this->is_multipage = az_is_set($index->options, 'multipage') && $index->perpage > 0;
        $this->ignorechars = az_get_ignorechars($index);
        
        $this->index = $index;
		$this->pageno = $pageno;
        $this->itemcache = unserialize($index->itemcache);
		
        if ($this->cache_disabled || $this->is_dirty()) {
			$this->build_index();
            $this->write_cache();
        } else {
            $this->build_index_from_cache();
		}
        $time_end = microtime(true);
        //az_println("Cache execution in: "+(($time_end - $time_start)*1000)."ms");
	}

    function build_index() {
    	$this->itemcache = 0;
        $this->items = $this->get_index_items($this->index);
        if ($this->is_multipage) {
            $this->pagecount = ceil(count($this->items) / $this->index->perpage);
            $this->pageno = ($this->pageno < 0) ? 0 : ($this->pageno >= $this->pagecount ? $this->pagecount - 1 : $this->pageno);
        }
        if ($this->has_alphalinks) {
            $this->alphalinks = $this->collect_links($this->items, $this->index->perpage);
        }
        $this->items = az_slice_array($this->items, $this->pageno, $this->pagecount, $this->index->perpage, $this->is_multipage);
	}

	function build_index_from_cache() {
        if ($this->has_alphalinks) {
            $this->alphalinks = unserialize($this->index->linkcache);
        }
        if ($this->is_multipage) {
            $this->pagecount = ceil(count($this->itemcache) / $this->index->perpage);
            $this->pageno = ($this->pageno < 0) ? 0 : ($this->pageno >= $this->pagecount ? $this->pagecount - 1 : $this->pageno);
        }
        $ids = az_slice_array($this->itemcache, $this->pageno, $this->pagecount, $this->index->perpage, $this->is_multipage);
        $this->items = $this->get_index_items($this->index, $ids);
	}	
	
    function write_cache() {
    	global $wpdb;
        
    	$id = $this->index->id;
    	
        if (!$this->cache_disabled) {
            $itemcache = serialize($this->itemcache);
            $linkcache = serialize($this->alphalinks);
        }
    	
        $query = "UPDATE ".AZ_TABLE." SET itemcache = '$itemcache', linkcache = '$linkcache' WHERE idindex = $id";
        $rc = $wpdb->query($query);
        //az_println("fn:write_cache : query = ".$query."<br/> rc = ".$rc);
    }

    function is_dirty() {
        $dirty = false;
        $idindex = $this->index->id;
        az_acquire_cache_mutex();
    	$dirtyposts = az_get_dirty_posts($idindex);
    	if (empty($this->itemcache)) {
    		$dirty = true;
            az_reset_dirty_posts($dirtyposts, $idindex);
    	} else if (!empty($dirtyposts)) {
            $dirty = $this->process_dirty_items($this->index, $this->itemcache, $dirtyposts[$idindex]);
            az_reset_dirty_posts($dirtyposts, $idindex);
    	}
        az_release_cache_mutex();
        //az_println("cache: ".(!empty($this->itemcache) ? implode(",", $this->itemcache) : "empty"));
        //az_println("dirty: ".(!empty($dirtyposts) ? implode(",", $dirtyposts[$idindex]) : "empty"));
        //az_println("cache is ".($dirty ? "dirty" : "clean"));
        return $dirty;
	}

	function process_dirty_items($index, $indexitems, $dirtyitems) {
        //az_trace("process dirty items: dirty items = ".implode(",", $dirtyitems));
		$dirty = false;
		foreach ($dirtyitems as $postid) {
            $pos = array_search($postid, $indexitems);
            if ($pos > 0) {
                $testhit[] = $indexitems[$pos - 1];
            }
            $testhit[] = $postid;                              
            if ($pos !== false && $pos < count($indexitems) - 1) { 
               $testhit[] = $indexitems[$pos + 1];
            }
            $result = $this->get_index_items($index, $testhit, true);
            $diff = array_diff_assoc($testhit, $result);
            //az_println("  test: ".implode(",", $testhit));
            //az_println("result: ".implode(",", $result));
            //az_println("  diff: ".implode(",", $diff));
            //az_println(" test1: ".((count($diff) > 0 && $pos !== false) ? "true" : "false"));
            //az_println(" test2: ".((count($diff) == 0 && $pos === false) ? "true" : "false"));
            if ((count($diff) > 0 && $pos !== false) || (count($diff) == 0 && $pos === false)) {
            	$dirty = true;
              	break;
            }
            unset($testhit);
		}
        return $dirty;
	}
	
    /**
     * Fetch the items to be added to the index from the blog's database.
     *
     * @param $index the parameters specificed for the index
     * @return items to be included in the index
     */
    function get_index_items($index, $cachedids = 0, $testsort = false) {
        global $wpdb;
        az_trace('fn:get_index_items: pageid: '.$post->ID);

        $fields = 'ID, post_title, post_excerpt, post_author';
        $query = "SELECT DISTINCT $fields FROM $wpdb->posts";
        $selection = "'post'";
        if (az_is_set($index->options, 'include-pages')) {
            if (az_is_set($index->options, 'include-pages-exclude-posts')) {
                $selection = "'page'";
            } else {
                $selection .= ",'page'";
            }
        }
        $where = "post_status = 'publish' AND post_type IN ($selection)";

        if (!empty($cachedids)) {
            $ids = implode(",", $cachedids);
        	$query .= " WHERE ID IN (".$ids.") AND $where ORDER BY FIELD(ID, $ids)";
        } else {
            if (!empty($index->tagids)) {
                $query .= " INNER JOIN $wpdb->term_relationships AS rel1 ON ($wpdb->posts.ID = rel1.object_id)"
                         ." INNER JOIN $wpdb->term_taxonomy AS tax1 ON (rel1.term_taxonomy_id = tax1.term_taxonomy_id"
                                                        ." AND tax1.taxonomy = 'post_tag' AND tax1.term_id IN ($index->tagids)) ";
            }
        
            if (!empty($index->catids)) {
                $query .= " INNER JOIN $wpdb->term_relationships AS rel2 ON ($wpdb->posts.ID = rel2.object_id)"
                         ." INNER JOIN $wpdb->term_taxonomy AS tax2 ON (rel2.term_taxonomy_id = tax2.term_taxonomy_id"
                                                        ." AND tax2.taxonomy = 'category' AND tax2.term_id IN ($index->catids))";
            }
            $query .= " WHERE $where";
        }
    
        $posts = $wpdb->get_results($query);
        //az_println("fn:get_index_items : query = ".$query."<br/> rc = ".$posts);
        
        // Build an array of items from the query results.
        $items = array();
        $hasfilter = has_filter('azindex_heading');
        foreach ($posts as $row) {
            $item = $this->get_post_index_info($index, $row, $hasfilter);
            if ($item != null) {
                $items[] = $item;
            }
        }                     
    
        // Sort the items into alphabetical order, as specified.
        if ($items && (empty($cachedids) || $testsort)) {
        	// Both globals only used in the compare function.
            global $az_indexchars, $az_nonalphaend;
            $az_indexchars = az_is_set($index->options, 'custom-links') ? $index->customlinks : AZ_INDEXCHARS;
            
            // Determine where to put the non-alpha starting entries by setting a signed variable.
            $az_nonalphaend = az_is_set($index->options, 'non-alpha-end') ? -1 : 1;
            
            // Sort the index using the specified comparison function.
            $comparefn = az_is_set($index->options, 'custom-sort') ? $index->customsort : 'az_compare';
            usort($items, $comparefn);
            
            if (!$this->cache_disabled) {
                if (!$testsort) {
                    // Add the ids of the sorted items to the item cache
                    unset($this->itemcache);
                    for ($i = 0; $i < count($items); $i++) {
                        $this->itemcache[] = $items[$i]['id'];
                    }
                } else {
                    // Just return the item list.
                    $justids = array();
                    for ($i = 0; $i < count($items); $i++) {
                        $justids[] = $items[$i]['id'];
                    }
                    $items = $justids;
                }
            }
        }
        return $items;    
    }

    function get_post_index_info($index, $row, $hasfilter = false) {
        $item = null;
        $head = trim($this->get_post_item($row->ID, $row->post_title, $row->post_excerpt, $row->post_author, $index->head, $index->headkey));
        if (!empty($head)) {
            if ($hasfilter) {
                $head = apply_filters('azindex_heading', $head);
            }
            // Remove any characters we want to ignore during the search.
            $sorthead = empty($this->ignorechars) ? $head : ltrim($head, $this->ignorechars);

            // Don't add item to the index if there is no heading.
            $item = array('id' => $row->ID,
                         'head' => $sorthead,
                         'fullhead' => $head, 
                         'subhead' => $this->get_post_item($row->ID, $row->post_title, $row->post_excerpt, $row->post_author, $index->subhead, $index->subheadkey),
                         'desc' => $this->get_post_item($row->ID, $row->post_title, $row->post_excerpt, $row->post_author, $index->desc, $index->desckey),
                        );
        }
        return $item;
    }
    
    /**
     * Get the item for a post specified in the index as the heading,
     * subheading, or description. 
     *
     * @param $type type of item to retrieve
     * @param $key item key, if the item is a custom field
     * @return item's value
     */
    function get_post_item($postid, $title, $excerpt, $author, $type, $key = false) {
        switch ($type) {
            case 'title':
                $item = $title;  
                break; 
            case 'excerpt':
                $item = $excerpt;
                break; 
            case 'author':
                $item = get_author_name($author);
                break; 
            case 'custom':
                // Only the first custom field found with key is used.
                $item = get_post_custom_values($key, $postid);
                $item = $item[0];
                break; 
        }   
        return $item;
    }

    /**
     * Collect the alphabetical links for the items in the index. Links are
     * created for the first item to begin with each letter.
     *
     * @param $items the items in the index
     * @param $perpage number of items per page
     * @return an array characters and what page of the index they are on 
    */
    function collect_links($items, $perpage) {
        for ($i = 0; $i < count($items); $i++) {
            $item = $items[$i];
            if (!empty($item)) {
                // Next entry starts with a different letter.
                if (strncasecmp($item['head'], $prevhead, 1)) {
                    $char = strtoupper($item['head'][0]);
                    $page = $this->is_multipage ? intval($i / $perpage) : 0;
                    $indexchars[] = array('char' => $char, 'page' => $page);
                }
            }
            $prevhead = $item['head'];
        }
        return $indexchars;
    }
}
/**
 * Compare function for sorting the index into alphabetical order, first using the 
 * heading, then the subheading, then the description.  Non-alpha numbers will be 
 * sorted to the top of the list (could make this an option at some point). 
 *
 * @param $in1 first index item to compare
 * @param $in2 second index item to compare
 * @return comparison result
 */
function az_compare($in1, $in2) {
    $rc = strcasecmp($in1['head'], $in2['head']);
    if ($rc == 0) {
        $rc = strcasecmp($in1['subhead'], $in2['subhead']);            
        if ($rc == 0) {
            $rc = strcasecmp($in1['desc'], $in2['desc']);            
        }
    } else {
        if (!empty($in1['head']) && !empty($in2['head'])) {
            global $az_indexchars, $az_nonalphaend;
            // Check to see if either heading starts with a non-alphanumeric character.
            $pos1 = strpos($az_indexchars, strtoupper($in1['head'][0]));
            $pos2 = strpos($az_indexchars, strtoupper($in2['head'][0]));            
            if ($pos1 === false && !($pos2 === false)) {
                $rc = -1 * $az_nonalphaend;
            } else if ($pos2 === false && !($pos1 === false)) {
                $rc = 1 * $az_nonalphaend;
            }
        }   
    }
    return $rc;
}

function az_slice_array($array, $pageno, $pagecount, $perpage, $multipage) {
        
    $start = 0;
    $length = count($array); 
    if ($multipage) {
        $length = $perpage;
        if ($pageno > 0) {
            // Keep the last item on the previous page in the front of the array if
            // we are not on the first (or only) page of the index. 
            $start = $pageno * $perpage - 1;
            $length++; 
        }
        if ($pageno < $pagecount - 1) {
            // Keep the first item on the next page at the end of the array if
            // we are not on the last (or only) page of the index. 
            $length++; 
        }
    }
    return array_slice($array, $start, $length);
}

function az_acquire_cache_mutex() {
    global $az_mutex;
    if (!isset($az_mutex)) {
        $az_mutex = new az_mutex(1, "az_sem_cache");
    }
    return $az_mutex->acquire();
}

function az_release_cache_mutex() {
    global $az_mutex;
    return $az_mutex->release();
}

class az_mutex {
    var $id;
    var $sem_id;
    var $is_acquired = false;
    var $use_flock = false;
    var $filename = '';
    var $filepointer;

    function az_mutex($id, $filename = '') {
        if (substr(PHP_OS, 0, 3) == 'WIN' || !function_exists('sem_get')) {
            $this->use_flock = true;
        }
        
        $this->id = $id;

        if ($this->use_flock) {
            if (empty($filename)){
                az_trace("Mutex:no filename specified");
            } else {
                $this->filename = $filename;
            }
        } else {
            if (!($this->sem_id = sem_get($this->id, 1))) {
                az_trace("Mutex:Error getting semaphore");
            }
        }
    }

    function acquire() {
        if ($this->use_flock) {
            if (($this->filepointer = @fopen($this->filename, "w+")) == false) {
                az_trace("Mutex:error opening mutex file<br>");
                return false;
            }
            if (flock($this->filepointer, LOCK_EX) == false) {
                az_trace("Mutex:error locking mutex file");
                return false;
            }
        } else {
            if (!sem_acquire($this->sem_id)) {
                az_trace("Mutex:error acquiring semaphore");
                return false;
            }
        }
        $this->is_acquired = true;
        return true;
    }

    function release() {
        if (!$this->is_acquired) {
            return true;
        }

        if ($this->use_flock) {
            if (flock($this->filepointer, LOCK_UN) == false) {
                az_trace("Mutex:error unlocking mutex file");
                return false;
            }
            fclose($this->filepointer);
        } else {
            if (!sem_release($this->sem_id)) {
                az_trace("Mutex:error releasing semaphore");
                return false;
            }
        }

        $this->is_acquired = false;
        return true;
    }

    function getId() {
        return $this->sem_id;
    }
}
?>